(*-------------------------------------------------------------------------- *)
(* Organize the libraries by theorems, types and constants. Give also        *)
(* additional information about the theorems such as their dependencies and  *)
(* originated theories (deph), their role (roleh) and the order (thmlo) in   *)
(* which the theorems were proved inside their theory.                       *)
(*-------------------------------------------------------------------------- *)

open Hh_parse
open Toolbox
open Preprocess

let no_split_flag = ref false 
let fof_flag = ref false
let thf_flag = ref false

(* dropping and getting functions *)
let drop_tt (_,s,r,t) = (s,r,t)
let drop_role (s,r,t) = (s,t)
let drop_term (s,r,t) = (s,r)
let obtain_name (s,r,t) = s

(* Transforming the lists of terms by theory *)
let flatten_tmll l = List.concat (List.map snd l)
let foreach_filter f tmll =
  List.map (fun (thy,l) -> (thy,List.filter f l)) tmll
let foreach_map f tmll = 
  List.map (fun (thy,l) -> (thy,List.map f l)) tmll

(* object partition *)
let rec is_type = function
  | Id "$t" -> true
  | Id "$type" -> true
  | Id "$tType" -> true
  | Id "$true" -> true
  | Comb (Comb (Id ">", Id "$t"), x) -> is_type x
  | Comb (Comb (Id ">", Id "$type"), x) -> is_type x
  | Comb (Comb (Id ">", Id "$tType"), x) -> is_type x
  | Comb (Comb (Id ">", Id "$true"), x) -> is_type x
  | Comb (Comb (Id "=>", Id "$t"), x) -> is_type x
  | Comb (Comb (Id "=>", Id "$type"), x) -> is_type x
  | Comb (Comb (Id "=>", Id "$tType"), x) -> is_type x
  | Comb (Comb (Id "=>", Id "$true"), x) -> is_type x
  | _ -> false;;

let is_type = function
  | Hh_parse.Comb (Hh_parse.Comb (Hh_parse.Id ":", x), y) -> is_type y
  | y -> is_type y;;

let rename_obj (s,r,t) = match t with
  | Hh_parse.Comb (Hh_parse.Comb (Hh_parse.Id ":", Id x), y) -> (x,r,y)
  | _ -> (s,r,t)

let is_thm_role role = 
  role = "ax" || role = "def" || role = "axiom" || role = "definition"
let is_thm (_,role,_) = is_thm_role role
let is_tyc (_,role,_) = role = "ty" || role = "type"
let is_ty (_,_,t) = is_type t

let is_thm_cj (_,role,_) =
  role = "ax" || role = "def" || role = "axiom" || role = "definition" ||
  role = "cj" || role = "conjecture"

(* Warning: here all non constants are assumed to be theorems *)
let get_tyl_cl_thml l = 
  let (tycl,thml) = List.partition is_tyc l in 
  let (tyl,cl) = List.partition is_ty tycl in
  let (tyl,cl,thml) = (List.map drop_role tyl, 
                       List.map drop_role cl, 
                       List.map drop_role thml) 
  in
  (tyl,cl,thml)

(* Ordering constants (obj here but should be changed to c) *)   
let get_cl objh t =
  let rl = ref [] in
  let rec loop t = match t with
  | Id x -> (if Hashtbl.mem objh x then rl := x :: !rl else ())
  | Comb (x, y) -> (loop x; loop y)
  | Abs (v,ty,tm) -> (loop ty; loop tm)
  in
  loop t;
  mk_fast_set compare !rl

let order_objl objl = 
  let objh = hash_of_alist objl in
  let graph = List.map (fun (x,t) -> (x, get_cl objh t)) objl in
  let sortedl = topo_sort graph in
  List.map (fun x -> (x, Hashtbl.find objh x)) sortedl

(* Initialization *)
let print_info (dir,tyl,cl,thml,thmlo) =
  log_endline ("Loading objects from " ^ dir);
  log_endline ("- types: " ^ string_of_int (List.length tyl));
  log_endline ("- constants: " ^ string_of_int (List.length cl));
  log_endline ("- theorems: " ^ string_of_int (List.length thml));
  log_endline ("- theories: "^ string_of_int (List.length thmlo) ^ "\n")

let init_dir dir =
  let (deph,tmll) = Read.read_dir dir in
  let tmll = foreach_map drop_tt tmll in
  let tmll = foreach_map rename_obj tmll in
  let (tyl,cl,thml) = get_tyl_cl_thml (flatten_tmll tmll) in
  let tmll = foreach_filter (is_thm_cj) tmll in
  let thmlo = foreach_map obtain_name tmll in 
  let roleh = hash_of_alist (List.map drop_term (flatten_tmll tmll)) in
  print_info (dir,tyl,cl,thml,thmlo);
  ((tyl,order_objl cl,thml),(deph,roleh,thmlo))

(* conjecture are included if we want to translate a thf file to fof*)
let init_file file =
  let tmll = Read.read_file file in
  let tmll = foreach_map drop_tt tmll in
  let tmll = foreach_map rename_obj tmll in
  let (tyl,cl,thml) = get_tyl_cl_thml (flatten_tmll tmll) in
  let tmll = foreach_filter (is_thm_cj) tmll in
  let thmlo = foreach_map obtain_name tmll in 
  let roleh = hash_of_alist (List.map drop_term (flatten_tmll tmll)) in
  print_info (file,tyl,cl,thml,thmlo);
  ((tyl,order_objl cl,thml),(roleh,thmlo))

(* 
Init for matching:
Gives an additional prefix to all the objects in the directory
Replace $t by $true etc
*)
let rec replace_tyimp_aux t = match t with
  | Id ">" -> Id "=>"
  | Id "$t" -> Id "$true" 
  | Id "$tType" -> Id "$true"
  | Comb(x,y) -> Comb(replace_tyimp_aux x, replace_tyimp_aux y)
  | Abs(v,ty,tm) -> Abs(v, replace_tyimp_aux ty, replace_tyimp_aux tm)
  | _ -> t 

let replace_tyimp (s,t) = (s,replace_tyimp_aux t)

let replace_tyimp_allobj (a,b,c) = 
  (List.map replace_tyimp a, List.map replace_tyimp b, List.map replace_tyimp c)

let smaller_thm_flag = ref false
let big_thml_glob = ref []

let init_dir_matching dir =
  let (deph,tmll) = Read.read_dir dir in
  let tmll = foreach_map drop_tt tmll in 
  let tmll = foreach_map rename_obj tmll in
  (* split *)
  let (deph,tmll) = 
    if !no_split_flag 
      then (deph,tmll) 
      else split_lib (deph,tmll) 
  in
  (* *)
  let (tyl,cl,thml) = get_tyl_cl_thml (flatten_tmll tmll) in
  let big_thml = List.map fst (List.filter (fun (_,t) -> size_of t >= 1000) thml) in
  big_thml_glob := big_thml;
  let deph = remove_deph big_thml deph in
  let tmll = if !smaller_thm_flag 
             then foreach_filter (fun (s,r,t) -> is_thm_cj (s,r,t) && 
                  not (List.mem s big_thml)) tmll 
             else foreach_filter (is_thm_cj) tmll in
  let thmlo = foreach_map obtain_name tmll in 
  let roleh = hash_of_alist (List.map drop_term (flatten_tmll tmll)) in
  print_info (dir,tyl,cl,thml,thmlo);
  ((tyl,order_objl cl,thml),(deph,roleh,thmlo))


(*--------------------------------------------------------------------------
  HH Printer 
  -------------------------------------------------------------------------- *)

let rec comb_arguments formula = match formula with
  | Comb (f1,f2) -> let (oper,argl) = comb_arguments f1 in 
                           (oper, (argl @ [f2])) (* could be optimized *)
  | _          -> (formula,[])

let rec hh_to_fof formula = match formula with
  | Id v -> v
  | Comb(Id "!", Abs (v,ty,tm)) -> 
    "(" ^ "! " ^ "[" ^ v ^ "] : " ^ hh_to_fof tm ^ ")"
  | Comb(Id "?", Abs (v,ty,tm)) -> 
    "(" ^ "? " ^ "[" ^ v ^ "] : " ^ hh_to_fof tm ^ ")"
  | Comb(Comb(Id("&"),f1),f2) -> 
    "(" ^ hh_to_fof f1 ^ " & " ^ hh_to_fof f2 ^ ")"
  | Comb(Comb(Id("|"),f1),f2) -> 
    "(" ^ hh_to_fof f1 ^ " | " ^ hh_to_fof f2 ^ ")"
  | Comb(Comb(Id("=>"),f1),f2) -> 
    "(" ^ hh_to_fof f1 ^ " => " ^ hh_to_fof f2 ^ ")"
  | Comb(Id("~"),f1) -> 
    "(~ " ^ hh_to_fof f1 ^ ")"
  | Comb(Comb(Id("="),f1),f2) -> 
    "(" ^ hh_to_fof f1 ^ " = " ^ hh_to_fof f2 ^ ")"
  | Comb(Comb(Id("<=>"),f1),f2) -> 
    "(" ^ hh_to_fof f1 ^ " <=> " ^ hh_to_fof f2 ^ ")"
  | Abs (v,ty,tm) -> "(" ^ "^ " ^ "[" ^ v ^ "] : " ^ hh_to_fof tm ^ ")"
  | Comb (x,y)    -> 
      let (oper, argl) = comb_arguments formula in
      hh_to_fof oper ^ "(" ^ String.concat "," (List.map hh_to_fof argl) ^ ")"

let rec comb_arguments formula = match formula with
  | Comb (f1,f2) -> let (oper,argl) = comb_arguments f1 in 
                           (oper, (argl @ [f2])) (* could be optimized *)
  | _          -> (formula,[])

let rec hh_to_tt formula = match formula with
  | Id v -> v
  | Comb(Id "!", Abs (v,ty,tm)) -> 
    "(" ^ "! " ^ "[" ^ v ^ ": " ^ hh_to_tt ty ^ "] : " ^ hh_to_tt tm ^ ")"
  | Comb(Id "?", Abs (v,ty,tm)) -> 
    "(" ^ "? " ^ "[" ^ v ^ ": " ^ hh_to_tt ty ^ "] : " ^ hh_to_tt tm ^ ")"
  | Comb(Comb(Id("&"),f1),f2) -> 
    "(" ^ hh_to_tt f1 ^ " & " ^ hh_to_tt f2 ^ ")"
  | Comb(Comb(Id("|"),f1),f2) -> 
    "(" ^ hh_to_tt f1 ^ " | " ^ hh_to_tt f2 ^ ")"
  | Comb(Comb(Id("=>"),f1),f2) -> 
    "(" ^ hh_to_tt f1 ^ " => " ^ hh_to_tt f2 ^ ")"
  | Comb(Id("~"),f1) -> 
    "(~ " ^ hh_to_tt f1 ^ ")"
  | Comb(Comb(Id("="),f1),f2) -> 
    "(" ^ hh_to_tt f1 ^ " = " ^ hh_to_tt f2 ^ ")"
  | Comb(Comb(Id("<=>"),f1),f2) -> 
    "(" ^ hh_to_tt f1 ^ " <=> " ^ hh_to_tt f2 ^ ")"
  | Abs (v,ty,tm) -> 
    "(" ^ "^ " ^ "[" ^ v ^ ": " ^ hh_to_tt ty ^ "] : " ^ hh_to_tt tm ^ ")"
  | Comb (x,y)    -> "(" ^ hh_to_tt x ^  " " ^ hh_to_tt y ^ ")"

(* Warning: quantifiers for types are different !> *)
let rec hh_to_thf formula = match formula with
  | Id v -> v
  | Comb(Id "!", Abs (v,ty,tm)) -> 
    "(" ^ "! " ^ "[" ^ v ^ ": " ^ hh_to_thf ty ^ "] : " ^ hh_to_thf tm ^ ")"
  | Comb(Id "?", Abs (v,ty,tm)) -> 
    "(" ^ "? " ^ "[" ^ v ^ ": " ^ hh_to_thf ty ^ "] : " ^ hh_to_thf tm ^ ")"
  | Comb(Comb(Id("&"),f1),f2) -> 
    "(" ^ hh_to_thf f1 ^ " & " ^ hh_to_thf f2 ^ ")"
  | Comb(Comb(Id("|"),f1),f2) -> 
    "(" ^ hh_to_thf f1 ^ " | " ^ hh_to_thf f2 ^ ")"
  | Comb(Comb(Id("=>"),f1),f2) -> 
    "(" ^ hh_to_thf f1 ^ " => " ^ hh_to_thf f2 ^ ")"
  | Comb(Id("~"),f1) -> 
    "(~ " ^ hh_to_thf f1 ^ ")"
  | Comb(Comb(Id("="),f1),f2) -> 
    "(" ^ hh_to_thf f1 ^ " = " ^ hh_to_thf f2 ^ ")"
  | Comb(Comb(Id("<=>"),f1),f2) -> 
    "(" ^ hh_to_thf f1 ^ " <=> " ^ hh_to_thf f2 ^ ")"
  | Abs (v,ty,tm) -> 
    "(" ^ "^ " ^ "[" ^ v ^ ": " ^ hh_to_thf ty ^ "] : " ^ hh_to_thf tm ^ ")"
  | Comb (x,y)    -> "(" ^ hh_to_thf x ^  " @ " ^ hh_to_thf y ^ ")"


(* 
 rlwrap ./top -I hh1
Init.init_dir (directory of the file should be .p );;

let palibs_dir = "../../hh2-data/palibs"
let coq_dir = palibs_dir ^ "/" ^ "Coq-matching";;
let h4_dir  = palibs_dir ^ "/" ^ "h4-kananaskis10/standard_library";;
let hl_dir  = palibs_dir ^ "/" ^ "hl-225/standard_library";;
let isa_dir = palibs_dir ^ "/" ^ "isa";;
let mat_dir = palibs_dir ^ "/" ^ "matita/matita_out_new";;
let miz_dir = palibs_dir ^ "/" ^ "mizar/xml2_out_bigger";;
let robot_dir = palibs_dir ^ "/" ^ "robot_benchmark";;
*)
