(*-------------------------------------------------------------------------- *)
(* Create a list of substitutions from a set of matching pair of constants.  *)
(* Rank them according to their scores.                                      *)
(*-------------------------------------------------------------------------- *)

open Init
open Toolbox 
open Pattern
open Matching_multiple
open Multiple_experiments

(*-------------------------------------------------------------------------- 
  Globals
  -------------------------------------------------------------------------- *)

let thm_domain_hash = Hashtbl.create max_thms
let thm_substl_hash = Hashtbl.create max_thms
let c_nodes_hash = Hashtbl.create max_constants

let eval_glob = ref []
let mix_flag = ref false

let init_eval_glob file = 
  let l0 = readl file in
  let l1 = List.map (List.hd ++ (Str.split (Str.regexp ":"))) l0 in
  eval_glob := l1

(*-------------------------------------------------------------------------- 
  Scoring
  -------------------------------------------------------------------------- *)

let sum_positive l = sum (List.filter (fun x -> x > 0.) l)

let aver_score_of_subst l = 
  sum_positive (List.map (Hashtbl.find ccscore_hash) l) /. 
    float_of_int (List.length l)

let aver_depscore_of_subst l =
  let ll = cartesian_product l l in
  sum_positive (List.map (find_dep_coeff dependency_coeff_hash) ll) /. 
    float_of_int (List.length ll)

let score_of_subst l =      
  let fl1 = try aver_score_of_subst l with _ -> 0. in
  let fl2 = try aver_depscore_of_subst l with _ -> 0. in
  fl1 *. fl2

(*-------------------------------------------------------------------------- 
  Printing
  -------------------------------------------------------------------------- *)

let string_of_redres (c1,c2) = 
  (name_of_cint inv_constant_hash1 c1) ^ " - " ^
  (name_of_cint inv_constant_hash2 c2)

let string_of_subst l = String.concat " , " (List.map string_of_redres l)

let string_of_domain domain = 
  String.concat " , " (List.map (name_of_cint inv_constant_hash1) domain)

(*-------------------------------------------------------------------------- 
  Substitution tools
  -------------------------------------------------------------------------- *)

let norm_substl substl =
  List.map (List.sort compare_fst) substl

let init_c_nodes_hash () =
  Hashtbl.clear c_nodes_hash;
  let iter (c1,c2) score = 
    let old_cl = (try Hashtbl.find c_nodes_hash c1 with _ -> []) in
    Hashtbl.replace c_nodes_hash c1 (((c1,c2),score) :: old_cl)
  in
  Hashtbl.iter iter ccscore_hash;
  let f_best c ccs = List.sort compare_score ccs in
  hash_map f_best c_nodes_hash

(*-------------------------------------------------------------------------- 
  Substitution creation for one theorem
  -------------------------------------------------------------------------- *)

let rec cartesian_productl ll = match ll with
    [] -> []
  | [l] -> List.map (fun x -> [x]) l
  | l :: m -> List.concat (List.map (fun m_el -> List.map (fun x -> x :: m_el) l) 
    (cartesian_productl m)) 

let rec length_of_ll ll = 
   if ll = [] 
   then (1,true)
   else 
     let (n,b) = length_of_ll (List.tl ll) in
     let n' = List.length (List.hd ll) * n in
     if (not b) || n' > 800 
      then (0,false)
      else (n',true)

let remove_one ll_ref =
  let compare_length a b = compare (List.length b) (List.length a) in
  ll_ref := List.sort compare_length !ll_ref;
  ll_ref := List.tl (List.hd !ll_ref) :: (List.tl !ll_ref)

let mix_substl substl =
  if substl = [] then [] else
  begin
  let linit = ref (List.map (fun (x,s) -> ((x,s),1.)) substl) in
  let lres = ref [] in
  let simil subst1 subst2 =
    float_of_int (List.length (list_inter subst1 subst2))
  in
  while (List.length !lres) < 50 && !linit <> [] do
    let l' = List.map (fun ((x,s),d) -> ((x,s), s /. (d *. d *. d) )) !linit in
    let best = fst (fst (List.hd (List.sort compare_score l'))) in
    lres := best :: !lres;
    linit := List.filter (fun x -> fst (fst x) <> best) !linit;
    linit := List.map ( fun ((x,s),d) -> ((x,s), d +. simil x best) ) !linit;
  done;
  List.rev !lres  
  end 
 
let create_substl thm = 
  let domain = try Hashtbl.find thm_domain_hash thm with _ -> [] in
  log_endline thm;
  log_endline ("  " ^ string_of_domain domain);
  let nodes1_of c = (try List.map fst (Hashtbl.find c_nodes_hash c) with _ -> []) in
  let is_in_past thm (_,c) =
    time_of_constant inv_constant_hash2 c <= time_of_thm thm
  in
  let nodes2_of c = 
    List.rev (List.filter (is_in_past thm) (nodes1_of c))
  in (* easier to remove worst score if they are at the top *)
  let nodes_list = ref (List.map nodes2_of domain) in
  let rec minimize () = 
    if snd (length_of_ll !nodes_list) = false
    then (remove_one nodes_list; minimize ())
    else ()
  in
  minimize ();
  let l1 = List.map (fun x -> (x, score_of_subst x)) (cartesian_productl !nodes_list) in
  let substl = first_n 200 (List.sort compare_score l1) in
  if !mix_flag 
    then norm_substl (mix_substl substl) 
    else norm_substl (List.map fst substl)


let init_thm_substl_hash thml =
  hash_init_of_keylist thm_substl_hash create_substl thml

let init_thm_domain_hash thml =
  Hashtbl.clear thm_domain_hash;
  let iter thm =
    let cl0 = snd (Hashtbl.find tp_hash1 thm) in
    (* mk_set because of properties not recomputed after equivalence closure *)
    let cl1 = mk_set (List.sort compare (deref cl0)) in 
    let cl2 = List.filter (Hashtbl.mem c_nodes_hash) cl1 in
    if cl2 <> [] then Hashtbl.add thm_domain_hash thm cl2 else ()
  in
  List.iter iter thml

let add_variant_glob thm = 
  if Hashtbl.mem thm_variant_hash1 thm 
    then eval_glob := Hashtbl.find thm_variant_hash1 thm :: !eval_glob
    else ()


let init_substl () = 
  (* directory *)
  let eval_thml = 
    if not (!eval_glob = [])
      then (eval_glob := List.filter (Hashtbl.mem tp_hash1) !eval_glob;
            let thml = !eval_glob in
            List.iter add_variant_glob thml;
            !eval_glob
           )
      else keys_of_hash tp_hash1
  in
  time "init_c_nodes_hash" init_c_nodes_hash ();
  time "init_thm_domain_hash" init_thm_domain_hash eval_thml;
  let n_thm = string_of_int (Hashtbl.length thm_domain_hash) in
  log_endline ("Number of initial theorems: " ^ n_thm);
  time "init_dependency_coeff_hash" init_dependency_coeff_hash ();
  time "init_thm_substl_hash" init_thm_substl_hash eval_thml;
  log_only_endline (string_of_int !time_of_thm_error ^ " time_of_thm warnings");
  log_only_endline (string_of_int !time_of_constant_error ^ " time_of_constant warnings")

let init_substl_all prover = 
  erase_file "log";
  Init.no_split_flag := true;
  Pattern.no_subterm_congruence_flag := true;
  time "self_matching: " self_exp prover;
  Init.no_split_flag := false;
  Pattern.no_subterm_congruence_flag := false;
  init_substl ()

let init_substl_all_dir dir_in dir_out = 
  erase_file "log";
  Init.no_split_flag := true;
  Pattern.no_subterm_congruence_flag := true;
  time "self_matching: " (self_exp_dir dir_in) dir_out;
  Init.no_split_flag := false;
  Pattern.no_subterm_congruence_flag := false;
  init_substl ()


(*  rlwrap ./top -I ../hh1 -I .. 
open Matching_subst;;
init_rev_substl_all prover
*)
