open Init
open Hh_parse
open Toolbox
open Read
open Pattern
open Matching_multiple
open Multiple_experiments
open Matching_subst

(*--------------------------------------------------------------------------
  Applying a substitution (not to variables)
  -------------------------------------------------------------------------- *)

let rec apply_subst_h subst_h formula = match formula with
  | _ when Hashtbl.mem subst_h formula -> Hashtbl.find subst_h formula
  | Id v -> Id v
  | Abs (v,ty,tm) -> 
    Abs(v, apply_subst_h subst_h ty, apply_subst_h subst_h tm)
  | Comb (x,y) -> 
    Comb(apply_subst_h subst_h x, apply_subst_h subst_h y)

let apply_named_subst subst formula = 
  apply_subst_h (hash_of_alist subst) formula

let rec mk_comb_list (oper,argl) = match argl with
  | []      -> oper
  | tm :: m -> mk_comb_list (Comb(oper,tm), m)

let rec subterm_hhm_to_hh ich formula = 
  match formula with
    C(c,ty)  -> Id (name_of_cint ich c)
  | App(f1,f2) -> 
     let (oper, argl) = fun_arguments formula in
     let (new_oper, new_argl) = 
       (subterm_hhm_to_hh ich oper, List.map (subterm_hhm_to_hh ich) argl)
     in
     mk_comb_list (new_oper, new_argl)
  | _ -> failwith "not valid subterm"

let apply_subst subst formula = 
  let f sh ich c =
    let name_c = name_of_cint ich c in
    if Hashtbl.mem sh name_c 
    then subterm_hhm_to_hh ich (Hashtbl.find sh name_c)
    else Id name_c
  in
  let f1f2 (c1,c2) = (f subterm_hash1 inv_constant_hash1 c1, 
                      f subterm_hash2 inv_constant_hash2 c2) 
  in 
  let hh_subst = List.filter (fun (x,y) -> x <> y) (List.map f1f2 subst) in
  apply_subst_h (hash_of_alist hh_subst) formula
 
(*--------------------------------------------------------------------------
  Rewriting all statements modulo equivalence classes
  -------------------------------------------------------------------------- *)

let distribute_cong l = 
  let f (namel,common_name) = List.map (fun x -> (Id x,Id common_name)) namel in
  List.concat (List.map f l)

let imatch_cong thml = (* each element is at most in one equiv class *)
  let named_subst = distribute_cong (!mem_intern_cong1) in
  List.map 
    (fun (name,formula) -> (name, apply_named_subst named_subst formula)) 
    thml

(*--------------------------------------------------------------------------
  Creating conjectures
  -------------------------------------------------------------------------- *)

let cjl_from_thm (name,formula) =    
  let substl = try Hashtbl.find thm_substl_hash name with _ -> [] in
  let variant_substl = 
    try 
      let variant_name = Hashtbl.find thm_variant_hash1 name in 
      Hashtbl.find thm_substl_hash variant_name 
    with _ -> []
  in
  let rec interleave l1 l2 = match (l1,l2) with 
    _,[] -> l1 
  | [],_ -> l2
  | (a :: m), (b :: m') -> a :: b :: interleave m m'
  in
  let f subst = 
    let new_formula = apply_subst subst formula in
    (name,subst,new_formula)
  in
  let third (a,b,c) = c in
  mk_set_fun third (List.map f (interleave substl variant_substl))

(*--------------------------------------------------------------------------
  Printing
  -------------------------------------------------------------------------- *)

(* statements *)
let string_of_thm (name,formula) = 
  if !fof_flag then
  "fof" ^ "(" ^ name ^ ", axiom, " ^ hh_to_fof formula ^ ")."
  else if !thf_flag then
  "thf" ^ "(" ^ name ^ ", axiom, " ^ "(" ^ name ^ " : " ^ 
  hh_to_thf formula ^ ")" ^ ")."  
  else
  "tt" ^ "(" ^ name ^ ", axiom, " ^ hh_to_tt formula ^ ")."

let write_thml_imatch file_out thml_imatch = 
  erase_file file_out;
  let iter thm = append file_out (string_of_thm thm ^ "\n") in
  List.iter iter thml_imatch

(* conjectures *)
let nb_cj_ref = ref 0

let string_of_cj (name,subst,formula) = 
  incr nb_cj_ref;
  let cj_name = 
    add_prime (remove_prime name ^ "__conjecture_" ^ string_of_int (!nb_cj_ref)) 
  in
  if !fof_flag then
  "fof" ^ "(" ^ cj_name ^ ", conjecture, " ^ hh_to_fof formula ^ ")."
  else if !thf_flag then
  "thf" ^ "(" ^ cj_name ^ ", conjecture, " ^ "(" ^ cj_name ^ " : " ^ 
  hh_to_thf formula ^ ")" ^ ")."  
  else
  "tt" ^ "(" ^ cj_name ^ ", conjecture, " ^ hh_to_tt formula ^ ")."

let string_of_cj_score (name,subst,formula) = 
  string_of_float (score_of_subst subst)

let string_of_cj_subst (name,subst,formula) = 
  string_of_subst subst

(* writing statistics *)
let write_cjl dir_out cjl = 
  if cjl = [] then ()
  else begin
  nb_cj_ref := 0;
  let (name,_,_) = List.hd cjl in
  let filename = Filename.basename (remove_prime name) in
  let l1 = List.map string_of_cj cjl in
  let l2 = List.map string_of_cj_score cjl in
  let l3 = List.map string_of_cj_subst cjl in
  mkdir_rec (dir_out ^ "/statements");
  mkdir_rec (dir_out ^ "/scores");
  mkdir_rec (dir_out ^ "/subst");
  writel (dir_out ^ "/statements/" ^ filename) l1;
  writel (dir_out ^ "/scores/" ^ filename) l2;
  writel (dir_out ^ "/subst/" ^ filename) l3
  end

(*--------------------------------------------------------------------------
  Main function
  -------------------------------------------------------------------------- *)

let create_conjecturel dir_out_prover = 
  let thml =
    if !eval_glob = [] then !thml1_init_glob 
    else List.filter (fun x -> List.mem (fst x) !eval_glob) !thml1_init_glob 
  in 
  let thml_imatch = imatch_cong thml in
  List.iter ((write_cjl (dir_out_prover ^ "/conjectures")) ++       
              cjl_from_thm) thml_imatch
  
let make_conjectures prover =
  let dir_out_prover = 
     if !subterm_flag 
       then data_dir ^ "/conjecturing/mix_subterm_exp/" ^ prover
       else data_dir ^ "/conjecturing/mix_exp/" ^ prover
  in
  log_file_glob := dir_out_prover ^ "/log";
  clean_dir dir_out_prover;
  mkdir_rec dir_out_prover;
  init_substl_all prover;
  time "conjecture_creation:" create_conjecturel dir_out_prover

let make_conjectures_dir dir_in dir_out =
  log_file_glob := dir_out ^ "/log";
  mkdir_rec dir_out;
  init_substl_all_dir dir_in dir_out;
  time "conjecture_creation:" create_conjecturel dir_out


(*  rlwrap ./top -I ../hh1 -I .. 
open Matching_translate;;
make_conjectures "h4";;
*)

(*  rlwrap ./top -I ../hh1 -I .. 
open Matching_translate;;
make_conjectures "mat";;

*)

(*  rlwrap ./top -I ../hh1 -I .. 
open Matching_translate;;
Pattern.subterm_flag := true;;
make_conjectures "mat";;
*)

(*  rlwrap ./top -I ../hh1 -I .. 
mix_flag := true;;
open Matching_translate;;
Pattern.init_time_hash "symbol_sequence";;
Matching_subst.init_eval_glob "hard";;
!Matching_subst.eval_glob;;
Toolbox.alist_of_hash (Pattern.time_constant_hash);;
Toolbox.alist_of_hash (Pattern.time_thm_hash);;
make_conjectures "miz_cj";;
*)

(*  rlwrap ./top -I ../hh1 -I .. 

mix_flag := true;;
open Matching_translate;;
Pattern.subterm_flag := true;;
Pattern.init_time_hash "symbol_sequence";;
Matching_subst.init_eval_glob "hard";;
!Matching_subst.eval_glob;;
Toolbox.alist_of_hash (Pattern.time_constant_hash);;
Toolbox.alist_of_hash (Pattern.time_thm_hash);;
make_conjectures "miz_cj";;
*)


